<?php

namespace InSegment\ApiCore\Models;

use InSegment\ApiCore\Traits\EventsReplace;
use InSegment\ApiCore\Traits\CachesAttributes;

use Illuminate\Support\Facades\DB;
use Illuminate\Support\Facades\Schema;
use Illuminate\Database\Eloquent\Model;

abstract class SupplementaryTable extends Model
{
    use EventsReplace;
    use CachesAttributes;
    
    /**
     * Indicate there is no incrementing key
     * 
     * @var bool 
     */
    public $incrementing = false;

    /**
     * The model has timestamps
     * 
     * @var bool
     */
    public $timestamps = false;
    
    /**
     * Run the migrations.
     *
     * @param array $options
     * @return null
     */
    public abstract static function establish(array $options);

    /**
     * Reverse the migrations.
     *
     * @param string|null $uid
     * @return null
     */
    public static function disband($uid = null)
    {
        Schema::dropIfExists((new static([], $uid))->getTable());
    }
    
    /**
     * Get query optionally specifying $uid
     * 
     * @param string|null $uid
     * @return \Illuminate\Database\Eloquent\Builder
     */
    public static function query($uid = null)
    {
        return (new static([], $uid))->newQuery();
    }
 
    /**
     * Check that the supplementary table exists
     * 
     * @param string|null $uid
     * @return bool
     */
    protected static function existsSupplementary($uid = null): bool
    {
        list($schema, $table) = explode('.', (new static([], $uid))->getTable());
        $selection = DB::select("select * from `information_schema`.`tables` where `table_schema` = ? and `table_name` = ?", [$schema, $table]);
        return count($selection) > 0;
    }
    
    /**
     * Execute create table statement with specified options and fields
     * 
     * @param array $options [
     *     'uid' => string|null,
     *     'inMemory' => bool|null,
     *     'temporary' => bool|null
     * ]
     * @param array $fields
     * @return null
     */
    protected static function createSupplementary(array $options, array $fields)
    {
        list($schema, $table) = explode('.', (new static([], $options['uid'] ?? null))->getTable());
        $typeOfTable = ($options['temporary'] ?? false) ? 'TEMPORARY TABLE' : 'TABLE';
        $engine = ($options['inMemory'] ?? false) ? 'MEMORY' : 'InnoDB';
        
        $implodeStatements = implode(",\n  ", $fields);
        
        DB::statement(
<<<SQL
CREATE {$typeOfTable} `{$schema}`.`{$table}` (
  {$implodeStatements}
) ENGINE={$engine} DEFAULT CHARSET=utf8
SQL
        );
    }
}
