<?php

namespace InSegment\ApiCore\Services;

use Illuminate\Support\Facades\DB;
use Illuminate\Support\Facades\Cache;

class ParseCreateTable
{
    /**
     * An instance
     * 
     * @var \InSegment\ApiCore\Services\ParseCreateTable 
     */
    private static $instance = null;

    /**
     * An array holding show creates for tables
     * 
     * @var array 
     */
    private $showCreateData;
    
    /**
     * Prefix of service data in cache
     * 
     * @var string
     */
    private $cachePrefix;
    
    /**
     * Lists column unique data of loaded table
     * 
     * @param string $table
     * @return string[][] - list of lists of columns of unique keys
     */
    public function getColumnUniqueData(string $table)
    {
        if (!isset($this->showCreateData['columnUnique'][$table])) {
            $this->loadColumnData($table);
        }
        
        return $this->showCreateData['columnUnique'][$table];
    }
    
    /**
     * Lists primary keys of loaded table
     * 
     * @param string $table
     * @return [
     *     'columns' => string[] - list of columns the primary key contains
     *     'isAi' => bool        - is PK an Auto-increment one
     * ] | null                  - null if the table has no PK
     */
    public function getPrimaryKeyData(string $table)
    {
        if (!isset($this->showCreateData['primaryKeys'][$table])) {
            $this->loadColumnData($table);
        }
        
        return $this->showCreateData['primaryKeys'][$table];
    }
    
    /**
     * Get auto_increment PRIMARY key
     * 
     * @param string $table
     * @return string|null
     */
    public function getAutoKey(string $table)
    {
        $pk = $this->getPrimaryKeyData($table);

        if ($pk !== null && $pk['isAi']) {
            return $pk['columns'][0];
        }
    }
    
    /**
     * Get query to create temporary table
     * 
     * If it is insert table, AUTO_INCREMENT key will be replaced with bigint(20) suitable for UUID_SHORT()
     * CONSTRAINT's are removed from query
     * 
     * @param string $table
     * @param array $uuidableTableMentions
     * @param bool $uuidable
     * @param bool $allowInMemory
     * @return array
     */
    public function &getCreateForTemp(string $table, array $uuidableTableMentions, bool $uuidable, bool $allowInMemory)
    {
        $this->loadShowCreate($table);
        
        $showCreate = $this->showCreateData['showCreate'][$table];
        $hasMentions = false;
        $last = count($showCreate) - 1;
        
        foreach ($showCreate as $i => $part) {
            $trimmed = trim($part);
            
            if ($trimmed[0] === '`') {
                if ($uuidableTableMentions && ($keyName = $this->mentionsMatches($trimmed, $uuidableTableMentions))) {
                    $hasMentions = true;
                    $showCreate[$i] = "`{$keyName}` bigint(20) unsigned NULL,";
                } else if ($allowInMemory && preg_match('/` ((tiny|medium|long)?(text|blob)|json)[ ,\(]/i', $trimmed)) {
                    $allowInMemory = false;
                }
            } else if (starts_with($trimmed, 'CONSTRAINT')) {
                unset($showCreate[$i]);
                
                if ($last == $i + 1) {
                    for ($j = $i; $j > 0; --$j) {
                        if (isset($showCreate[$j])) {
                            $showCreate[$j] = rtrim($showCreate[$j], ',');
                            break;
                        }
                    }
                }
            }
        }
        
        if ($allowInMemory) {
            $showCreate[$last] = str_replace('ENGINE=InnoDB', 'ENGINE=MEMORY', $showCreate[$last]);
        }
        
        if ($uuidable && $hasMentions) {
            $showCreate[$last] = preg_replace('/AUTO_INCREMENT=\d+/', '', $showCreate[$last], 1);
        }

        return $showCreate;
    }
    
    /**
     * Sets the data about the primary key and unique key columns of the table
     * 
     * @param string $table
     * @return $this
     */
    public function loadColumnData(string $table)
    {
        if (!isset($this->showCreateData['columnUnique'][$table])) {
            $this->loadShowCreate($table);
            $cacheKey = "{$this->cachePrefix}.column-unique.{$table}";
            $cachedColumnData = Cache::rememberForever(
                $cacheKey, function () use ($table) {
                    $columnUniqueForTable = [];
                    $primaryKeyInfo = null;
                    $showCreate = $this->showCreateData['showCreate'][$table];
                    $createTable = array_shift($showCreate);
                    $engineAIAndCharset = array_pop($showCreate);
                    $aiColumnName = null;
                    
                    foreach ($showCreate as $tableStatement) {
                        $tableStatement = trim($tableStatement);
                        $primary = strpos($tableStatement, 'PRIMARY KEY') === 0;
                        $unique = !$primary && strpos($tableStatement, 'UNIQUE KEY') === 0;
                        
                        if ($tableStatement[0] === '`' && strpos($tableStatement, 'AUTO_INCREMENT') > -1) {
                            $aiColumnName = substr($tableStatement, 1, strpos($tableStatement, '`', 1) - 1);
                        }
                        
                        if ($primary || $unique) {
                            $matches = [];
                            if (preg_match('/\(`([^\)]+)`\)/', $tableStatement, $matches)) {
                                $matchedColumnName = explode('`,`', $matches[1]);
                                $columnUniqueForTable[] = $matchedColumnName;
                                if ($primary) {
                                    $primaryKeyInfo = [
                                        'columns' => $matchedColumnName,
                                        'isAi' => in_array($aiColumnName, $matchedColumnName)
                                    ];
                                }
                            }
                        }
                    }
                    
                    return ['primary' => $primaryKeyInfo, 'unique' => $columnUniqueForTable];
                }
            );
            
            $this->showCreateData['primaryKeys'][$table] = $cachedColumnData['primary'];
            $this->showCreateData['columnUnique'][$table] = $cachedColumnData['unique'];
        }
        
        return $this;
    }
    
    /**
     * Sets the data about table creation specifics
     * 
     * @param string $table
     * @return $this
     */
    protected function loadShowCreate(string $table)
    {
        if (!isset($this->showCreateData['showCreate'][$table])) {
            $this->showCreateData['showCreate'][$table] = Cache::rememberForever(
                "{$this->cachePrefix}.show-create.{$table}", function () use ($table) {
                    return explode("\n", DB::selectOne("SHOW CREATE TABLE `{$table}`")->{'Create Table'});
                }
            );
        }
        
        return $this;
    }
    
    /**
     * Test for uuidable mentions match, remove matched from further matching
     * 
     * @param string $trimmed
     * @param array $mentions
     * @return bool|string
     */
    protected function mentionsMatches(string $trimmed, array &$mentions)
    {
        $matches = [];
        if (preg_match('/^`(' . implode('|', $mentions) . ')`/', $trimmed, $matches)) {
            unset($mentions[$matches[1]]);
            return $matches[1];
        }
        
        return false;
    }
    
    /**
     * Get ParseCreateTable instance
     * 
     * @return \InSegment\ApiCore\Services\ParseCreateTable
     */
    public static function getInstance()
    {
        if (self::$instance == null) {
            self::$instance = new ParseCreateTable();
        }
        
        return self::$instance;
    }
    
    /**
     * A singleton, no instantiation from outside
     */
    private function __construct()
    {
        $this->showCreateData = [
            'primaryKeys' => [],
            'columnUnique' => [],
            'showCreate' => []
        ];
        
        $this->cachePrefix = config('api_core.cache_prefix', 'insegment.api-core.') . 'show-create-data';
    }

}
